package gihyo_sample

import (
	"fmt"
	"math/rand"
	"sync"
	"time"
)

type PingPongPayload struct {
	Counter int
}

func ExamplePingPong() {
	var p PingPongPayload
	chA := make(chan *PingPongPayload)
	chB := make(chan *PingPongPayload)

	var wg sync.WaitGroup
	wg.Add(2)
	go func() {
		defer wg.Done()
		for {
			p, ok := <-chA // chAから読み込み
			if !ok {
				break
			}
			fmt.Printf("chA: p.Counter = %d\n", p.Counter)
			p.Counter++ // データの内容を書き換え
			if p.Counter > 6 {
				break
			}

			chB <- p
		}
		close(chB) // chBへ書き込み
	}()

	go func() {
		defer wg.Done()
		for {
			p, ok := <-chB // chBから読み込み
			if !ok {
				break
			}
			fmt.Printf("chB: p.Counter = %d\n", p.Counter)
			p.Counter++ // データの内容を書き換え
			if p.Counter > 6 {
				break
			}

			chA <- p // chAへ書き込み
		}
		close(chA)
	}()

	chA <- &p // chA へ最初の書き込み
	wg.Wait() // 終了を待つ

	// Output:
	// chA: p.Counter = 0
	// chB: p.Counter = 1
	// chA: p.Counter = 2
	// chB: p.Counter = 3
	// chA: p.Counter = 4
	// chB: p.Counter = 5
	// chA: p.Counter = 6
}

func ExampleBufferedChannel() {
	ch := make(chan int, 5)
	go func() {
		defer close(ch)
		for i := 0; i < 10; i++ {
			fmt.Println("Writing", i)
			ch <- i
		}
	}()

	time.Sleep(2 * time.Second)

	for {
		v, ok := <-ch
		if !ok {
			break
		}
		fmt.Println("Read", v)
	}

	// Unordered output:
	// Writing 0
	// Writing 1
	// Writing 2
	// Writing 3
	// Writing 4
	// Writing 5
	// Read 0
	// Read 1
	// Writing 6
	// Read 2
	// Writing 7
	// Writing 8
	// Writing 9
	// Read 3
	// Read 4
	// Read 5
	// Read 6
	// Read 7
	// Read 8
	// Read 9
}

func ExampleCloseOnBlockingChannelRead() {
	ch := make(chan struct{})
	var wg sync.WaitGroup

	for i := 0; i < 10; i++ {
		wg.Add(1)
		go func(i int) {
			defer wg.Done()
			defer fmt.Printf("goroutine %d exiting\n", i)

			// channelが閉じられるまでブロックする
			<-ch
		}(i)
	}

	// 5秒後にchannelを閉じる
	time.AfterFunc(5*time.Second, func() {
		fmt.Println("5 seconds elapsed, closing channel")
		close(ch)
	})

	wg.Wait()
	fmt.Println("done")
	// Unordered output:
	// 5 seconds elapsed, closing channel
	// goroutine 0 exiting
	// goroutine 1 exiting
	// goroutine 2 exiting
	// goroutine 3 exiting
	// goroutine 4 exiting
	// goroutine 5 exiting
	// goroutine 6 exiting
	// goroutine 7 exiting
	// goroutine 8 exiting
	// goroutine 9 exiting
	// done
}

func ExampleSelectOnNilChannel() {
	ch1 := make(chan struct{})
	ch2 := make(chan struct{})
	done := make(chan struct{})

	go func() {
		defer close(done)
		for {
			select {
			case <-ch1:
				fmt.Println("ch1")
			case _, ok := <-ch2:
				if !ok {
					return
				}
				fmt.Println("ch2")
				ch1 = nil
			}
		}
	}()

	time.AfterFunc(time.Second, func() {
		ch2 <- struct{}{}
	})
	time.AfterFunc(5*time.Second, func() {
		close(ch2)
	})

	<-done
	fmt.Println("done")
	// Output:
	// ch2
	// done
}

func ExampleReadClosedChannel() {
	ch := make(chan int, 10)
	for i := 1; i <= 10; i++ {
		ch <- i
	}
	close(ch)

	for i := 0; i < 10; i++ {
		v := <-ch // 1から10まで読み込まれる
		fmt.Println("Read", v)
	}
	v := <-ch // i = 0. intのゼロ値は0
	fmt.Println("Read", v)
	// Output:
	// Read 1
	// Read 2
	// Read 3
	// Read 4
	// Read 5
	// Read 6
	// Read 7
	// Read 8
	// Read 9
	// Read 10
	// Read 0
}

func ExampleWaitGroup() {
	var wg sync.WaitGroup
	for i := 0; i < 10; i++ {
		wg.Add(1)
		go func() {
			defer wg.Done()
			n := rand.Intn(5)
			fmt.Println("Going to sleep for", n, "seconds")
			time.Sleep(time.Duration(n) * time.Second)
		}()
	}

	wg.Wait() // 終了を待つ
	fmt.Println("All goroutines finished")
	// Unordered output:
	// Going to sleep for 1 seconds
	// Going to sleep for 2 seconds
	// Going to sleep for 2 seconds
	// Going to sleep for 1 seconds
	// Going to sleep for 3 seconds
	// Going to sleep for 0 seconds
	// Going to sleep for 0 seconds
	// Going to sleep for 0 seconds
	// Going to sleep for 1 seconds
	// Going to sleep for 4 seconds
	// All goroutines finished
}

func ExampleCond() {
	var wg sync.WaitGroup
	var mu sync.Mutex
	c := sync.NewCond(&mu)

	for i := 0; i < 10; i++ {
		wg.Add(1)
		go func(i int) {
			defer wg.Done()
			defer mu.Unlock()

			fmt.Printf("waiting %d\n", i)
			mu.Lock()
			c.Wait()
			fmt.Printf("go %d\n", i)
		}(i)
	}

	// 通知
	for i := 0; i < 10; i++ {
		time.Sleep(100 * time.Millisecond)
		fmt.Printf("signaling!\n")
		c.Signal()
	}

	wg.Wait()

	// Unordered output:
	// waiting 0
	// waiting 5
	// waiting 1
	// waiting 2
	// waiting 3
	// waiting 4
	// waiting 7
	// waiting 6
	// waiting 9
	// waiting 8
	// signaling!
	// go 0
	// signaling!
	// go 5
	// signaling!
	// go 1
	// signaling!
	// go 2
	// signaling!
	// go 3
	// signaling!
	// go 4
	// signaling!
	// go 7
	// signaling!
	// go 6
	// signaling!
	// go 9
	// signaling!
	// go 8
}

var initGlobal sync.Once

// 複数goroutineから呼ばれる可能性のある関数
func ExampleOnce() {
	cb := func(wg *sync.WaitGroup) {
		defer wg.Done()
		fmt.Println("callback called")
		initGlobal.Do(func() {
			// Globalの初期化…
			fmt.Println("initializing")
		})
	}
	var wg sync.WaitGroup
	for i := 0; i < 10; i++ {
		wg.Add(1)
		go cb(&wg)
	}
	wg.Wait()

	// Output:
	// callback called
	// initializing
	// callback called
	// callback called
	// callback called
	// callback called
	// callback called
	// callback called
	// callback called
	// callback called
	// callback called
}

func ExampleMutex() {
	var o ExampleMutexObject
	o.NeedsLock()
}

type ExampleMutexObject struct {
	mu sync.Mutex
}

func (o *ExampleMutexObject) NeedsLock() {
	o.mu.Lock()
	defer o.mu.Unlock()

	fmt.Println("Locked: NeedsLock")
}

func ExampleRWMutex() {
	var mu sync.RWMutex
	var wg sync.WaitGroup

	for i := 1; i <= 5; i++ {
		wg.Add(1)
		id := i
		go func() {
			defer wg.Done()
			for i := 0; i < 5; i++ {
				// 共有ロックを得る
				mu.RLock()
				fmt.Printf("Reader %d: Acquired lock\n", id)
				time.Sleep(time.Second)
				mu.RUnlock()
			}
		}()
	}

	time.AfterFunc(3*time.Second, func() {
		fmt.Println("Writer: Acquired lock")
		// 排他ロックを得る
		mu.Lock()
	})
	time.AfterFunc(6*time.Second, func() {
		fmt.Println("Writer: Releasing lock")
		// 排他ロックを解放する
		mu.Unlock()
	})

	wg.Wait()
	// Unordered output:
	//Reader 3: Acquired lock
	//Reader 4: Acquired lock
	//Reader 5: Acquired lock
	//Reader 1: Acquired lock
	//Reader 2: Acquired lock
	//Reader 3: Acquired lock
	//Reader 4: Acquired lock
	//Reader 5: Acquired lock
	//Reader 1: Acquired lock
	//Reader 2: Acquired lock
	//Reader 1: Acquired lock
	//Reader 5: Acquired lock
	//Reader 4: Acquired lock
	//Reader 2: Acquired lock
	//Reader 3: Acquired lock
	//Writer: Acquired lock
	//Writer: Releasing lock
	//Reader 3: Acquired lock
	//Reader 1: Acquired lock
	//Reader 5: Acquired lock
	//Reader 4: Acquired lock
	//Reader 2: Acquired lock
	//Reader 2: Acquired lock
	//Reader 1: Acquired lock
	//Reader 5: Acquired lock
	//Reader 4: Acquired lock
	//Reader 3: Acquired lock
}
