package gihyo_sample

import (
	"context"
	"errors"
	"fmt"
	"log"
	"net/http"
	"sync"
	"time"
)

func FetchURL(sem chan struct{}, u string) {
	sem <- struct{}{}
	defer func() { <-sem }()

	_, _ = http.Get(u)
	fmt.Println("fetched", u)
}

func ExampleSemaphore() {
	sem := make(chan struct{}, 10)
	urls := []string{
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
		"http://gihyo.jp",
	}

	var wg sync.WaitGroup
	for _, u := range urls {
		wg.Add(1)
		go func() {
			defer wg.Done()
			FetchURL(sem, u)
		}()
	}

	wg.Wait()

	// Output:
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
	// fetched http://gihyo.jp
}

func GenerateSeries(max int) <-chan int {
	c := make(chan int)
	go func() {
		defer close(c)
		for i := 0; i < max; i++ {
			c <- i
		}
	}()
	return c
}

func ReadSeries(wg *sync.WaitGroup, c <-chan int) {
	defer wg.Done()
	for i := range c {
		fmt.Println("Read", i)
	}
}

func ExampleGenerator() {
	c := GenerateSeries(100)

	var wg sync.WaitGroup

	wg.Add(2)
	go ReadSeries(&wg, c)
	go ReadSeries(&wg, c)

	wg.Wait()

	// Unordered output:
	// Read 1
	// Read 0
	// Read 2
	// Read 4
	// Read 5
	// Read 6
	// Read 3
	// Read 8
	// Read 9
	// Read 7
	// Read 10
	// Read 11
	// Read 12
	// Read 14
	// Read 15
	// Read 16
	// Read 17
	// Read 18
	// Read 19
	// Read 20
	// Read 21
	// Read 13
	// Read 23
	// Read 24
	// Read 25
	// Read 26
	// Read 27
	// Read 28
	// Read 22
	// Read 29
	// Read 31
	// Read 32
	// Read 30
	// Read 34
	// Read 35
	// Read 36
	// Read 37
	// Read 33
	// Read 38
	// Read 40
	// Read 41
	// Read 39
	// Read 42
	// Read 43
	// Read 45
	// Read 46
	// Read 47
	// Read 48
	// Read 49
	// Read 44
	// Read 50
	// Read 52
	// Read 53
	// Read 51
	// Read 55
	// Read 56
	// Read 57
	// Read 58
	// Read 54
	// Read 59
	// Read 61
	// Read 62
	// Read 60
	// Read 63
	// Read 64
	// Read 66
	// Read 67
	// Read 68
	// Read 69
	// Read 70
	// Read 65
	// Read 71
	// Read 73
	// Read 74
	// Read 75
	// Read 76
	// Read 72
	// Read 77
	// Read 78
	// Read 80
	// Read 81
	// Read 82
	// Read 83
	// Read 84
	// Read 79
	// Read 85
	// Read 87
	// Read 88
	// Read 86
	// Read 89
	// Read 90
	// Read 92
	// Read 93
	// Read 91
	// Read 94
	// Read 95
	// Read 97
	// Read 98
	// Read 99
	// Read 96
}

func ExampleTimer() {
	t := time.NewTimer(time.Second)
	defer t.Stop()

	for {
		select {
		case <-t.C:
			break
		default:
			fmt.Println("default")
		}
	}
}

func ExampleTicker() {
	t := time.NewTicker(time.Second)
	defer t.Stop()

	for i := 1; i <= 10; i++ {
		select {
		case <-t.C:
			fmt.Printf("%d\n", i)
		}
	}

	// Output:
	// 1
	// 2
	// 3
	// 4
	// 5
	// 6
	// 7
	// 8
	// 9
	// 10
}

func ExampleSelectTicker() {
	t := time.NewTicker(time.Second)
	defer t.Stop()

	ctx, cancel := context.WithCancel(context.Background())

	in := make(chan struct{})
	go func() {
		defer cancel()
		for i := 0; i < 10; i++ {
			in <- struct{}{}
		}
	}()

	var c chan struct{}
	for {
		select {
		case <-ctx.Done():
			return
		case <-t.C:
			c = in
		case <-c:
			c = nil
			fmt.Println("new item!")
		}
	}

	// Output:
	// new item!
	// new item!
	// new item!
	// new item!
	// new item!
	// new item!
	// new item!
	// new item!
	// new item!
	// new item!
}

func ExampleContextRelationship() {
	ctx1, cancel1 := context.WithCancel(context.Background())
	ctx2, cancel2 := context.WithCancel(ctx1)
	var wg sync.WaitGroup

	wg.Add(2)
	go func() {
		defer wg.Done()
		log.Printf("waiting for ctx1...")
		<-ctx1.Done()
		log.Printf("ctx1.Done() returns")
	}()

	go func() {
		defer wg.Done()
		log.Printf("waiting for ctx2...")
		<-ctx2.Done()
		log.Printf("ctx2.Done() returns")
	}()

	time.AfterFunc(time.Second*5, cancel2) // A
	time.AfterFunc(time.Second*1, cancel1) // B

	wg.Wait()
}

func WorkWithContext(ctx context.Context) {
	for {
		select {
		case <-ctx.Done():
			return
			// default: // この例では割愛
			// Doneが返ってこなければキャンセルされていないので
			// 本来の処理を行う。
		}
	}
}

func ExampleContext() {
	// キャンセル処理可能なContextを作成
	ctx, cancel := context.WithCancel(context.Background())
	// この関数が終わったら必ずキャンセルする
	defer cancel()

	// 処理を行うgoroutine
	var wg sync.WaitGroup
	wg.Add(1)
	go func() {
		defer wg.Done()
		WorkWithContext(ctx)
	}()

	time.AfterFunc(5*time.Second, cancel)

	wg.Wait()
}

type FanoutTask struct{}

var FanoutTaskCount = 0 // サンプル用のなんちゃってカウンタ
func FanoutGetNextTask() (FanoutTask, error) {
	FanoutTaskCount++
	if FanoutTaskCount > 10 {
		return FanoutTask{}, errors.New("no more tasks")
	}

	return FanoutTask{}, nil
}

func FanoutDispatcher(out chan FanoutTask) {
	defer close(out)
	// DBとかからタスクを取得していく
	for {
		task, err := FanoutGetNextTask()
		if err != nil {
			return
		}

		// channelに書き込む
		out <- task
	}
}

func FanoutWorker(in chan FanoutTask) {
	for {
		_, ok := <-in
		if !ok {
			return
		}
		fmt.Println("Process task")
	}
}

func ExampleFanout() {
	ch := make(chan FanoutTask)

	var wg sync.WaitGroup
	for i := 0; i < 10; i++ {
		wg.Add(1)
		go func() {
			defer wg.Done()
			FanoutWorker(ch)
		}()
	}

	FanoutDispatcher(ch)

	wg.Wait()

	// Output:
	// Process task
	// Process task
	// Process task
	// Process task
	// Process task
	// Process task
	// Process task
	// Process task
	// Process task
	// Process task
}
